##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.admin.disableInviteCodes`.

    Attributes:
        codes: Codes.
        accounts: Accounts.
    """

    accounts: Optional[List[str]] = None
    codes: Optional[List[str]] = None



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional, Union

from atproto import CID
from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Commit(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`.

    Attributes:
        seq: Seq.
        rebase: Rebase.
        tooBig: Too big.
        repo: Repo.
        commit: Commit.
        prev: Prev.
        blocks: CAR file containing relevant blocks.
        ops: Ops.
        blobs: Blobs.
        time: Time.
    """

    blobs: List[CID]
    blocks: Union[str, bytes]
    commit: CID
    ops: List['models.ComAtprotoSyncSubscribeRepos.RepoOp']
    rebase: bool
    repo: str
    seq: int
    time: str
    tooBig: bool
    prev: Optional[CID] = None

    _type: str = 'com.atproto.sync.subscribeRepos#commit'


@dataclass
class Handle(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`.

    Attributes:
        seq: Seq.
        did: Did.
        handle: Handle.
        time: Time.
    """

    did: str
    handle: str
    seq: int
    time: str

    _type: str = 'com.atproto.sync.subscribeRepos#handle'


@dataclass
class Migrate(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`.

    Attributes:
        seq: Seq.
        did: Did.
        migrateTo: Migrate to.
        time: Time.
    """

    did: str
    seq: int
    time: str
    migrateTo: Optional[str] = None

    _type: str = 'com.atproto.sync.subscribeRepos#migrate'


@dataclass
class Tombstone(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`.

    Attributes:
        seq: Seq.
        did: Did.
        time: Time.
    """

    did: str
    seq: int
    time: str

    _type: str = 'com.atproto.sync.subscribeRepos#tombstone'


@dataclass
class Info(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`.

    Attributes:
        name: Name.
        message: Message.
    """

    name: str
    message: Optional[str] = None

    _type: str = 'com.atproto.sync.subscribeRepos#info'


@dataclass
class RepoOp(base.ModelBase):

    """Definition model for :obj:`com.atproto.sync.subscribeRepos`.

    Attributes:
        action: Action.
        path: Path.
        cid: Cid.
    """

    action: str
    path: str
    cid: Optional[CID] = None

    _type: str = 'com.atproto.sync.subscribeRepos#repoOp'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client.models import base


@dataclass
class Label(base.ModelBase):

    """Definition model for :obj:`com.atproto.label.defs`. Metadata tag on an atproto resource (eg, repo or record)

    Attributes:
        src: DID of the actor who created this label.
        uri: AT URI of the record, repository (account), or other resource which this label applies to.
        cid: optionally, CID specifying the specific version of 'uri' resource this label applies to.
        val: the short string name of the value or type of this label.
        neg: if true, this is a negation label, overwriting a previous label.
        cts: timestamp when this label was created.
    """

    cts: str
    src: str
    uri: str
    val: str
    cid: Optional[str] = None
    neg: Optional[bool] = None

    _type: str = 'com.atproto.label.defs#label'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.graph.getList`.

    Attributes:
        list: List.
        limit: Limit.
        cursor: Cursor.
    """

    list: str
    cursor: Optional[str] = None
    limit: Optional[int] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.graph.getList`.

    Attributes:
        cursor: Cursor.
        list: List.
        items: Items.
    """

    items: List['models.AppBskyGraphDefs.ListItemView']
    list: 'models.AppBskyGraphDefs.ListView'
    cursor: Optional[str] = None



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from typing_extensions import Literal

ReasonType = Literal['ReasonSpam', 'ReasonViolation', 'ReasonMisleading', 'ReasonSexual', 'ReasonRude', 'ReasonOther']

ReasonSpam: Literal['reasonSpam'] = 'reasonSpam'

ReasonViolation: Literal['reasonViolation'] = 'reasonViolation'

ReasonMisleading: Literal['reasonMisleading'] = 'reasonMisleading'

ReasonSexual: Literal['reasonSexual'] = 'reasonSexual'

ReasonRude: Literal['reasonRude'] = 'reasonRude'

ReasonOther: Literal['reasonOther'] = 'reasonOther'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.feed.getPosts`.

    Attributes:
        uris: Uris.
    """

    uris: List[str]


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.feed.getPosts`.

    Attributes:
        posts: Posts.
    """

    posts: List['models.AppBskyFeedDefs.PostView']



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.notification.listNotifications`.

    Attributes:
        limit: Limit.
        cursor: Cursor.
        seenAt: Seen at.
    """

    cursor: Optional[str] = None
    limit: Optional[int] = None
    seenAt: Optional[str] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.notification.listNotifications`.

    Attributes:
        cursor: Cursor.
        notifications: Notifications.
    """

    notifications: List['models.AppBskyNotificationListNotifications.Notification']
    cursor: Optional[str] = None


@dataclass
class Notification(base.ModelBase):

    """Definition model for :obj:`app.bsky.notification.listNotifications`.

    Attributes:
        uri: Uri.
        cid: Cid.
        author: Author.
        reason: Expected values are 'like', 'repost', 'follow', 'mention', 'reply', and 'quote'.
        reasonSubject: Reason subject.
        record: Record.
        isRead: Is read.
        indexedAt: Indexed at.
        labels: Labels.
    """

    author: 'models.AppBskyActorDefs.ProfileView'
    cid: str
    indexedAt: str
    isRead: bool
    reason: str
    record: 'base.RecordModelBase'
    uri: str
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None
    reasonSubject: Optional[str] = None

    _type: str = 'app.bsky.notification.listNotifications#notification'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.repo.listRecords`.

    Attributes:
        repo: The handle or DID of the repo.
        collection: The NSID of the record type.
        limit: The number of records to return.
        cursor: Cursor.
        rkeyStart: DEPRECATED: The lowest sort-ordered rkey to start from (exclusive).
        rkeyEnd: DEPRECATED: The highest sort-ordered rkey to stop at (exclusive).
        reverse: Reverse the order of the returned records?
    """

    collection: str
    repo: str
    cursor: Optional[str] = None
    limit: Optional[int] = None
    reverse: Optional[bool] = None
    rkeyEnd: Optional[str] = None
    rkeyStart: Optional[str] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.repo.listRecords`.

    Attributes:
        cursor: Cursor.
        records: Records.
    """

    records: List['models.ComAtprotoRepoListRecords.Record']
    cursor: Optional[str] = None


@dataclass
class Record(base.ModelBase):

    """Definition model for :obj:`com.atproto.repo.listRecords`.

    Attributes:
        uri: Uri.
        cid: Cid.
        value: Value.
    """

    cid: str
    uri: str
    value: 'base.RecordModelBase'

    _type: str = 'com.atproto.repo.listRecords#record'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.listAppPasswords`.

    Attributes:
        passwords: Passwords.
    """

    passwords: List['models.ComAtprotoServerListAppPasswords.AppPassword']


@dataclass
class AppPassword(base.ModelBase):

    """Definition model for :obj:`com.atproto.server.listAppPasswords`.

    Attributes:
        name: Name.
        createdAt: Created at.
    """

    createdAt: str
    name: str

    _type: str = 'com.atproto.server.listAppPasswords#appPassword'



from atproto.xrpc_client import models

RECORD_TYPE_TO_MODEL_CLASS = {
    'app.bsky.actor.profile': models.AppBskyActorProfile.Main,
    'app.bsky.feed.repost': models.AppBskyFeedRepost.Main,
    'app.bsky.graph.listitem': models.AppBskyGraphListitem.Main,
    'app.bsky.feed.like': models.AppBskyFeedLike.Main,
    'app.bsky.graph.follow': models.AppBskyGraphFollow.Main,
    'app.bsky.graph.list': models.AppBskyGraphList.Main,
    'app.bsky.graph.block': models.AppBskyGraphBlock.Main,
    'app.bsky.feed.post': models.AppBskyFeedPost.Main,
}



from atproto.cid import CID


class BlobRef:
    def __init__(self, blob_type: str, mime_type: str, ref: str, size: int):
        self.blob_type = blob_type
        self.mime_type = mime_type
        self.ref = ref
        self.size = size

    @property
    def cid(self) -> CID:
        return CID.decode(self.ref)

    @classmethod
    def from_dict(cls, data: dict) -> 'BlobRef':
        return cls(
            blob_type=data['$type'],
            mime_type=data['mimeType'],
            ref=data['ref']['$link'],
            size=int(data['size']),
        )

    def to_dict(self) -> dict:
        return {'$type': self.blob_type, 'mimeType': self.mime_type, 'size': self.size, 'ref': {'$link': self.ref}}



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from typing_extensions import Literal

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class ListViewBasic(base.ModelBase):

    """Definition model for :obj:`app.bsky.graph.defs`.

    Attributes:
        uri: Uri.
        name: Name.
        purpose: Purpose.
        avatar: Avatar.
        viewer: Viewer.
        indexedAt: Indexed at.
    """

    name: str
    purpose: 'models.AppBskyGraphDefs.ListPurpose'
    uri: str
    avatar: Optional[str] = None
    indexedAt: Optional[str] = None
    viewer: Optional['models.AppBskyGraphDefs.ListViewerState'] = None

    _type: str = 'app.bsky.graph.defs#listViewBasic'


@dataclass
class ListView(base.ModelBase):

    """Definition model for :obj:`app.bsky.graph.defs`.

    Attributes:
        uri: Uri.
        creator: Creator.
        name: Name.
        purpose: Purpose.
        description: Description.
        descriptionFacets: Description facets.
        avatar: Avatar.
        viewer: Viewer.
        indexedAt: Indexed at.
    """

    creator: 'models.AppBskyActorDefs.ProfileView'
    indexedAt: str
    name: str
    purpose: 'models.AppBskyGraphDefs.ListPurpose'
    uri: str
    avatar: Optional[str] = None
    description: Optional[str] = None
    descriptionFacets: Optional[List['models.AppBskyRichtextFacet.Main']] = None
    viewer: Optional['models.AppBskyGraphDefs.ListViewerState'] = None

    _type: str = 'app.bsky.graph.defs#listView'


@dataclass
class ListItemView(base.ModelBase):

    """Definition model for :obj:`app.bsky.graph.defs`.

    Attributes:
        subject: Subject.
    """

    subject: 'models.AppBskyActorDefs.ProfileView'

    _type: str = 'app.bsky.graph.defs#listItemView'


ListPurpose = Literal['Modlist']

Modlist: Literal['modlist'] = 'modlist'


@dataclass
class ListViewerState(base.ModelBase):

    """Definition model for :obj:`app.bsky.graph.defs`.

    Attributes:
        muted: Muted.
    """

    muted: Optional[bool] = None

    _type: str = 'app.bsky.graph.defs#listViewerState'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.server.createInviteCode`.

    Attributes:
        useCount: Use count.
        forAccount: For account.
    """

    useCount: int
    forAccount: Optional[str] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.createInviteCode`.

    Attributes:
        code: Code.
    """

    code: str



import os
from enum import Enum
from pathlib import Path
from typing import Tuple, Union

from atproto.codegen import (
    DISCLAIMER,
    INPUT_MODEL,
    OUTPUT_MODEL,
    PARAMS_MODEL,
    append_code,
    capitalize_first_symbol,
    format_code,
    gen_description_by_camel_case_name,
)
from atproto.codegen import get_code_intent as _
from atproto.codegen import get_file_path_parts, get_import_path, join_code, write_code
from atproto.codegen.models import builder
from atproto.exceptions import InvalidNsidError
from atproto.lexicon import models
from atproto.nsid import NSID

_MODELS_OUTPUT_DIR = Path(__file__).parent.parent.parent.joinpath('xrpc_client', 'models')


class ModelType(Enum):
    PARAMS = 'Parameters'
    DATA = 'Input data'
    RESPONSE = 'Output data'
    DEF = 'Definition'
    RECORD = 'Record'


def get_def_model_name(method_name: str) -> str:
    return f'{capitalize_first_symbol(method_name)}'


def get_model_path(nsid: NSID, method_name: str) -> str:
    return f'models.{get_import_path(nsid)}.{get_def_model_name(method_name)}'


def save_code(nsid: NSID, code: str) -> None:
    path_to_file = _MODELS_OUTPUT_DIR.joinpath(*get_file_path_parts(nsid))
    write_code(_MODELS_OUTPUT_DIR.joinpath(path_to_file), code)


def save_code_part(nsid: NSID, code: str) -> None:
    path_to_file = _MODELS_OUTPUT_DIR.joinpath(*get_file_path_parts(nsid))
    append_code(_MODELS_OUTPUT_DIR.joinpath(path_to_file), code)


def _get_model_imports() -> str:
    # TODO(MarshalX): isort can't delete unused imports. mb add ruff
    lines = [
        'from dataclasses import dataclass',
        'from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union',
        '',
        'from typing_extensions import Literal',
        'from atproto.xrpc_client import models',
        'from atproto.xrpc_client.models import base',
        'from atproto.xrpc_client.models.blob_ref import BlobRef',
        '',
        'from atproto import CID',
        '',
    ]

    return join_code(lines)


_NSID_WITH_IMPORTS = set()


def _save_code_import_if_not_exist(nsid) -> None:
    if nsid not in _NSID_WITH_IMPORTS:
        lines = [DISCLAIMER, _get_model_imports()]
        save_code(nsid, join_code(lines))
        _NSID_WITH_IMPORTS.add(nsid)


def _get_model_class_def(name: str, model_type: ModelType) -> str:
    lines = ['@dataclass']

    if model_type is ModelType.PARAMS:
        lines.append(f'class {PARAMS_MODEL}(base.ParamsModelBase):')
    elif model_type is ModelType.DATA:
        lines.append(f'class {INPUT_MODEL}(base.DataModelBase):')
    elif model_type is ModelType.RESPONSE:
        lines.append(f'class {OUTPUT_MODEL}(base.ResponseModelBase):')
    elif model_type is ModelType.RECORD:
        lines.append(f'class {get_def_model_name(name)}(base.RecordModelBase):')
    elif model_type is ModelType.DEF:
        lines.append(f'class {get_def_model_name(name)}(base.ModelBase):')

    lines.append('')

    return join_code(lines)


_LEXICON_TYPE_TO_PRIMITIVE_TYPEHINT = {
    models.LexString: 'str',
    models.LexInteger: 'int',
    models.LexBoolean: 'bool',
}


def _get_optional_typehint(type_hint, *, optional: bool) -> str:
    if optional:
        return f'Optional[{type_hint}] = None'
    else:
        return type_hint


def _get_ref_typehint(nsid: NSID, field_type_def, *, optional: bool) -> str:
    model_path, _ = _resolve_nsid_ref(nsid, field_type_def.ref)
    return _get_optional_typehint(f"'{model_path}'", optional=optional)


def _resolve_nsid_ref(nsid: NSID, ref: str, *, local: bool = False) -> Tuple[str, str]:
    """Returns path to the model and model name"""
    if '#' in ref:
        ref_nsid_str, def_name = ref.split('#', 1)
        def_name = get_def_model_name(def_name)

        try:
            ref_nsid = NSID.from_str(ref_nsid_str)
            return get_model_path(ref_nsid, def_name), def_name
        except InvalidNsidError:
            if local:
                return def_name, def_name
            return get_model_path(nsid, def_name), def_name
    else:
        ref_nsid = NSID.from_str(ref)
        def_name = get_def_model_name(nsid.name)

        if local:
            return def_name, def_name

        # FIXME(MarshalX): Is it works well? ;d
        return get_model_path(ref_nsid, 'Main'), def_name
        # return get_model_path(ref_nsid, ref_nsid.name), def_name


def _get_ref_union_typehint(nsid: NSID, field_type_def, *, optional: bool) -> str:
    def_names = []
    for ref in field_type_def.refs:
        import_path, _ = _resolve_nsid_ref(nsid, ref)
        def_names.append(import_path)

    # unbelievable but it's true. If schema doesn't describe the right type in Union
    # we should fall back to the plain data
    # maybe it's for the records that have custom fields... idk
    # ref: https://github.com/bluesky-social/atproto/blob/b01e47b61730d05a780f7a42667b91ccaa192e8e/packages/lex-cli/src/codegen/lex-gen.ts#L325
    # grep by "{$type: string; [k: string]: unknown}" string
    def_names.append('Dict[str, Any]')

    def_names = ', '.join([f"'{name}'" for name in def_names])
    return _get_optional_typehint(f"Union[{def_names}]", optional=optional)


def _get_model_field_typehint(nsid: NSID, field_name: str, field_type_def, *, optional: bool) -> str:
    field_type = type(field_type_def)

    if field_type == models.LexUnknown:
        # TODO(MarshalX): some of "unknown" types are well known...
        return _get_optional_typehint(f"'base.RecordModelBase'", optional=optional)

    type_hint = _LEXICON_TYPE_TO_PRIMITIVE_TYPEHINT.get(field_type)
    if type_hint:
        return _get_optional_typehint(type_hint, optional=optional)

    if field_type is models.LexArray:
        items_type_hint = _get_model_field_typehint(nsid, field_name, field_type_def.items, optional=False)
        return _get_optional_typehint(f'List[{items_type_hint}]', optional=optional)

    if field_type is models.LexRef:
        return _get_ref_typehint(nsid, field_type_def, optional=optional)

    if field_type is models.LexRefUnion:
        return _get_ref_union_typehint(nsid, field_type_def, optional=optional)

    if field_type is models.LexCidLink:
        return _get_optional_typehint(f'CID', optional=optional)

    if field_type is models.LexBytes:
        # CAR file containing relevant blocks
        return _get_optional_typehint('Union[str, bytes]', optional=optional)

    if field_type is models.LexBlob:
        # yes, it returns blob,but actually it's blob ref here
        return _get_optional_typehint('BlobRef', optional=optional)

    raise ValueError(f'Unknown field type {field_name.__name__}')


def _get_req_fields_set(lex_obj: Union[models.LexObject, models.LexXrpcParameters]) -> set:
    required_fields = set()
    if lex_obj.required:
        required_fields = set(lex_obj.required)

    if hasattr(lex_obj, 'nullable') and lex_obj.nullable:
        # TODO(MarshalX): not 100% the same thing. think about it
        required_fields = required_fields.difference(set(lex_obj.nullable))

    return required_fields


def _get_model_docstring(
    nsid: Union[str, NSID], lex_object: Union[models.LexObject, models.LexXrpcParameters], model_type: ModelType
) -> str:
    model_desc = lex_object.description or ''
    model_desc = f"{model_type.value} model for :obj:`{nsid}`. {model_desc}"

    doc_string = [f'{_(1)}"""{model_desc}', '', f'{_(1)}Attributes:']

    for field_name, field_type in lex_object.properties.items():
        field_desc = field_type.description
        if field_desc is None:
            field_desc = gen_description_by_camel_case_name(field_name)
        if field_desc[-1] not in {'.', '?', '!'}:
            field_desc += '.'

        doc_string.append(f'{_(2)}{field_name}: {field_desc}')

    doc_string.append(f'{_(1)}"""')
    doc_string.append('')

    return join_code(doc_string)


def _get_model(nsid: NSID, lex_object: Union[models.LexObject, models.LexXrpcParameters]) -> str:
    required_fields = _get_req_fields_set(lex_object)

    fields = []
    optional_fields = []

    for field_name, field_type in lex_object.properties.items():
        is_optional = field_name not in required_fields
        type_hint = _get_model_field_typehint(nsid, field_name, field_type, optional=is_optional)

        field_def = f'{_(1)}{field_name}: {type_hint}'  # TODO(MarshalX): Add default values. for bool, etc..
        if is_optional:
            optional_fields.append(field_def)
        else:
            fields.append(field_def)

    optional_fields.sort()
    fields.sort()

    fields.extend(optional_fields)

    fields.append('')

    return join_code(fields)


def _get_model_ref(nsid: NSID, ref: models.LexRef) -> str:
    # FIXME(MarshalX): "local=True" Is it works well? ;d
    ref_class, _ = _resolve_nsid_ref(nsid, ref.ref, local=True)
    ref_typehint = f'Type[{ref_class}]'

    # "Ref" suffix required to fix name collisions from different namespaces
    lines = [
        f'#: {OUTPUT_MODEL} reference to :obj:`{ref_class}` model.',
        f'{OUTPUT_MODEL}Ref: {ref_typehint} = {ref_class}',
        '',
        '',
    ]

    return join_code(lines)


def _get_model_raw_data(name: str) -> str:
    lines = [f'#: {name} raw data type.', f'{name}: Union[Type[str], Type[bytes]] = bytes\n\n']
    return join_code(lines)


def _generate_params_model(nsid: NSID, definition: Union[models.LexXrpcProcedure, models.LexXrpcQuery]) -> str:
    lines = [_get_model_class_def(nsid.name, ModelType.PARAMS)]

    if definition.parameters:
        lines.append(_get_model_docstring(nsid, definition.parameters, ModelType.PARAMS))
        lines.append(_get_model(nsid, definition.parameters))

    return join_code(lines)


def _generate_xrpc_body_model(nsid: NSID, body: models.LexXrpcBody, model_type: ModelType) -> str:
    lines = []
    if body.schema:
        if isinstance(body.schema, models.LexObject):
            lines.append(_get_model_class_def(nsid.name, model_type))
            lines.append(_get_model_docstring(nsid, body.schema, model_type))
            lines.append(_get_model(nsid, body.schema))
    else:
        if model_type is ModelType.DATA:
            model_name = INPUT_MODEL
        elif model_type is ModelType.RESPONSE:
            model_name = OUTPUT_MODEL
        else:
            raise ValueError('Wrong type or not implemented')

        lines.append(_get_model_raw_data(model_name))

    return join_code(lines)


def _generate_data_model(nsid: NSID, input_body: models.LexXrpcBody) -> str:
    return _generate_xrpc_body_model(nsid, input_body, ModelType.DATA)


def _generate_response_model(nsid: NSID, output_body: models.LexXrpcBody) -> str:
    return _generate_xrpc_body_model(nsid, output_body, ModelType.RESPONSE)


def _generate_def_model(nsid: NSID, def_name: str, def_model: models.LexObject, model_type: ModelType) -> str:
    lines = [
        _get_model_class_def(def_name, model_type),
        _get_model_docstring(nsid, def_model, model_type),
        _get_model(nsid, def_model),
    ]

    def_type = f'{nsid}#{def_name}'
    if def_name == 'main':
        def_type = str(nsid)

    lines.append(f"{_(1)}_type: str = '{def_type}'")

    lines.append('')

    return join_code(lines)


def _generate_def_token(def_name: str) -> str:
    lines = [
        f"{get_def_model_name(def_name)}: Literal['{def_name}'] = '{def_name}'",
        '',
        '',
    ]
    return join_code(lines)


def _generate_def_array(nsid: NSID, def_name: str, def_model: models.LexArray) -> str:
    return f'{get_def_model_name(def_name)} = {_get_model_field_typehint(nsid, def_name, def_model, optional=False)}\n'


def _generate_def_string(def_name: str, def_model: models.LexString) -> str:
    # FIXME(MarshalX): Doesn't support all fields

    if not def_model.knownValues:
        return ''

    # FIXME(MarshalX): Use ref resolver
    known_values = ["'" + get_def_model_name(v.split('#', 1)[1]) + "'" for v in def_model.knownValues]
    known_values = ', '.join(known_values)

    type_ = f'Literal[{known_values}]'

    lines = [
        f"{get_def_model_name(def_name)} = {type_}",
        '',
        '',
    ]

    return join_code(lines)


def _generate_params_models(lex_db: builder.LexDB) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        definition = defs['main']
        if definition.parameters:
            save_code_part(nsid, _generate_params_model(nsid, definition))


def _generate_data_models(lex_db: builder.LexDB) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        definition = defs['main']
        if definition.input:
            save_code_part(nsid, _generate_data_model(nsid, definition.input))


def _generate_response_models(lex_db: builder.LexDB) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        definition = defs['main']
        if definition.output:
            save_code_part(nsid, _generate_response_model(nsid, definition.output))


def _generate_def_models(lex_db: builder.LexDB) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        for def_name, def_model in defs.items():
            if isinstance(def_model, models.LexToken):
                save_code_part(nsid, _generate_def_token(def_name))
            elif isinstance(def_model, models.LexString):
                save_code_part(nsid, _generate_def_string(def_name, def_model))
            elif isinstance(def_model, models.LexObject):
                save_code_part(nsid, _generate_def_model(nsid, def_name, def_model, ModelType.DEF))
            elif isinstance(def_model, models.LexArray):
                save_code_part(nsid, _generate_def_array(nsid, def_name, def_model))
            else:
                raise ValueError(f'Unhandled type {type(def_model)}')


def _generate_record_models(lex_db: builder.LexDB) -> None:
    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        for def_name, def_record in defs.items():
            if isinstance(def_record, models.LexRecord):
                # TODO(MarshalX): Process somehow def_record.key?
                save_code_part(nsid, _generate_def_model(nsid, def_name, def_record.record, ModelType.RECORD))


def _generate_record_type_database(lex_db: builder.LexDB) -> None:
    lines = ['from atproto.xrpc_client import models', 'RECORD_TYPE_TO_MODEL_CLASS = {']

    for nsid, defs in lex_db.items():
        _save_code_import_if_not_exist(nsid)

        for def_name, def_record in defs.items():
            # for now there are records only in under "main" definition name.
            # need to rework a bit if this behaviour will be changed
            if isinstance(def_record, models.LexRecord):
                class_name = get_def_model_name(def_name)
                record_type = str(nsid)

                path_to_class = f'models.{get_import_path(nsid)}.{class_name}'

                lines.append(f"'{record_type}': {path_to_class},")

    lines.append('}')

    write_code(_MODELS_OUTPUT_DIR.joinpath('type_conversion.py'), join_code(lines))


def _generate_ref_models(lex_db: builder.LexDB) -> None:
    for nsid, defs in lex_db.items():
        definition = defs['main']
        if hasattr(definition, 'input') and definition.input and definition.input.schema:
            if isinstance(definition.input.schema, models.LexRef):
                save_code_part(nsid, _get_model_ref(nsid, definition.input.schema))

        if hasattr(definition, 'output') and definition.output and definition.output.schema:
            if isinstance(definition.output.schema, models.LexRef):
                save_code_part(nsid, _get_model_ref(nsid, definition.output.schema))


def _generate_init_files(root_package_path: Path) -> None:
    # one of the ways that I tried. Doesn't work well due to circular imports
    for root, dirs, files in os.walk(root_package_path):
        root = Path(root)

        import_lines = []
        for dir_name in dirs:
            if dir_name.startswith('__'):
                continue

            import_parts = root.parts[root.joinpath(dir_name).parts.index('xrpc_client') :]
            from_import = '.'.join(import_parts)

            if dir_name in {'app', 'com'}:
                continue

            import_lines.append(f'from {from_import} import {dir_name}')

        for file_name in files:
            if file_name.startswith('__'):
                continue

            import_parts = root.parts[root.parts.index('xrpc_client') :]
            from_import = '.'.join(import_parts)

            import_lines.append(f'from atproto.{from_import} import {file_name[:-3]}')

        if root.name == 'models':
            # FIXME skip for now. should be generated too
            continue

        if root.name == '__pycache__':
            continue

        write_code(root.joinpath('__init__.py'), join_code(import_lines))


def _generate_empty_init_files(root_package_path: Path):
    for root, dirs, files in os.walk(root_package_path):
        root = Path(root)

        for dir_name in dirs:
            if dir_name.startswith('__'):
                continue

            if dir_name in {'app', 'com'}:
                continue

        for file_name in files:
            if file_name.startswith('__'):
                continue

        if root.name == 'models':
            # FIXME skip for now. should be generated too
            continue

        if root.name == '__pycache__':
            continue

        write_code(root.joinpath('__init__.py'), DISCLAIMER)


def _generate_import_aliases(root_package_path: Path):
    # is generates __init__.py file if models dir with aliases like this;
    # from xrpc_client.models.app.bsky.actor import defs as AppBskyActorDefs

    import_lines = []
    for root, dirs, files in os.walk(root_package_path):
        root = Path(root)

        if root == root_package_path:
            continue

        for file in files:
            if file.startswith('.') or file.startswith('__') or file.endswith('.pyc'):
                continue

            import_parts = root.parts[root.parts.index('xrpc_client') :]
            from_import = '.'.join(import_parts)

            nsid_parts = list(root.parts[root.parts.index('models') + 1 :]) + file[:-3].split('_')
            alias_name = ''.join([p.capitalize() for p in nsid_parts])

            import_lines.append(f'from atproto.{from_import} import {file[:-3]} as {alias_name}')

    write_code(_MODELS_OUTPUT_DIR.joinpath('__init__.py'), join_code(import_lines))


def generate_models():
    _generate_params_models(builder.build_params_models())
    _generate_data_models(builder.build_data_models())
    _generate_response_models(builder.build_response_models())
    _generate_def_models(builder.build_def_models())

    _generate_record_models(builder.build_record_models())
    _generate_record_type_database(builder.build_record_models())

    # refs should be generated at the end!
    _generate_ref_models(builder.build_refs_models())

    _generate_empty_init_files(_MODELS_OUTPUT_DIR)
    _generate_import_aliases(_MODELS_OUTPUT_DIR)

    format_code(_MODELS_OUTPUT_DIR)

    print('Done')


if __name__ == '__main__':
    generate_models()



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Labels(base.ModelBase):

    """Definition model for :obj:`com.atproto.label.subscribeLabels`.

    Attributes:
        seq: Seq.
        labels: Labels.
    """

    labels: List['models.ComAtprotoLabelDefs.Label']
    seq: int

    _type: str = 'com.atproto.label.subscribeLabels#labels'


@dataclass
class Info(base.ModelBase):

    """Definition model for :obj:`com.atproto.label.subscribeLabels`.

    Attributes:
        name: Name.
        message: Message.
    """

    name: str
    message: Optional[str] = None

    _type: str = 'com.atproto.label.subscribeLabels#info'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.describeServer`.

    Attributes:
        inviteCodeRequired: Invite code required.
        availableUserDomains: Available user domains.
        links: Links.
    """

    availableUserDomains: List[str]
    inviteCodeRequired: Optional[bool] = None
    links: Optional['models.ComAtprotoServerDescribeServer.Links'] = None


@dataclass
class Links(base.ModelBase):

    """Definition model for :obj:`com.atproto.server.describeServer`.

    Attributes:
        privacyPolicy: Privacy policy.
        termsOfService: Terms of service.
    """

    privacyPolicy: Optional[str] = None
    termsOfService: Optional[str] = None

    _type: str = 'com.atproto.server.describeServer#links'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.repo.deleteRecord`.

    Attributes:
        repo: The handle or DID of the repo.
        collection: The NSID of the record collection.
        rkey: The key of the record.
        swapRecord: Compare and swap with the previous record by cid.
        swapCommit: Compare and swap with the previous commit by cid.
    """

    collection: str
    repo: str
    rkey: str
    swapCommit: Optional[str] = None
    swapRecord: Optional[str] = None



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from typing_extensions import Literal

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class ActionView(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        id: Id.
        action: Action.
        subject: Subject.
        subjectBlobCids: Subject blob cids.
        createLabelVals: Create label vals.
        negateLabelVals: Negate label vals.
        reason: Reason.
        createdBy: Created by.
        createdAt: Created at.
        reversal: Reversal.
        resolvedReportIds: Resolved report ids.
    """

    action: 'models.ComAtprotoAdminDefs.ActionType'
    createdAt: str
    createdBy: str
    id: int
    reason: str
    resolvedReportIds: List[int]
    subject: Union['models.ComAtprotoAdminDefs.RepoRef', 'models.ComAtprotoRepoStrongRef.Main', 'Dict[str, Any]']
    subjectBlobCids: List[str]
    createLabelVals: Optional[List[str]] = None
    negateLabelVals: Optional[List[str]] = None
    reversal: Optional['models.ComAtprotoAdminDefs.ActionReversal'] = None

    _type: str = 'com.atproto.admin.defs#actionView'


@dataclass
class ActionViewDetail(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        id: Id.
        action: Action.
        subject: Subject.
        subjectBlobs: Subject blobs.
        createLabelVals: Create label vals.
        negateLabelVals: Negate label vals.
        reason: Reason.
        createdBy: Created by.
        createdAt: Created at.
        reversal: Reversal.
        resolvedReports: Resolved reports.
    """

    action: 'models.ComAtprotoAdminDefs.ActionType'
    createdAt: str
    createdBy: str
    id: int
    reason: str
    resolvedReports: List['models.ComAtprotoAdminDefs.ReportView']
    subject: Union[
        'models.ComAtprotoAdminDefs.RepoView',
        'models.ComAtprotoAdminDefs.RepoViewNotFound',
        'models.ComAtprotoAdminDefs.RecordView',
        'models.ComAtprotoAdminDefs.RecordViewNotFound',
        'Dict[str, Any]',
    ]
    subjectBlobs: List['models.ComAtprotoAdminDefs.BlobView']
    createLabelVals: Optional[List[str]] = None
    negateLabelVals: Optional[List[str]] = None
    reversal: Optional['models.ComAtprotoAdminDefs.ActionReversal'] = None

    _type: str = 'com.atproto.admin.defs#actionViewDetail'


@dataclass
class ActionViewCurrent(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        id: Id.
        action: Action.
    """

    action: 'models.ComAtprotoAdminDefs.ActionType'
    id: int

    _type: str = 'com.atproto.admin.defs#actionViewCurrent'


@dataclass
class ActionReversal(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        reason: Reason.
        createdBy: Created by.
        createdAt: Created at.
    """

    createdAt: str
    createdBy: str
    reason: str

    _type: str = 'com.atproto.admin.defs#actionReversal'


ActionType = Literal['Takedown', 'Flag', 'Acknowledge', 'Escalate']

Takedown: Literal['takedown'] = 'takedown'

Flag: Literal['flag'] = 'flag'

Acknowledge: Literal['acknowledge'] = 'acknowledge'

Escalate: Literal['escalate'] = 'escalate'


@dataclass
class ReportView(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        id: Id.
        reasonType: Reason type.
        reason: Reason.
        subject: Subject.
        reportedBy: Reported by.
        createdAt: Created at.
        resolvedByActionIds: Resolved by action ids.
    """

    createdAt: str
    id: int
    reasonType: 'models.ComAtprotoModerationDefs.ReasonType'
    reportedBy: str
    resolvedByActionIds: List[int]
    subject: Union['models.ComAtprotoAdminDefs.RepoRef', 'models.ComAtprotoRepoStrongRef.Main', 'Dict[str, Any]']
    reason: Optional[str] = None

    _type: str = 'com.atproto.admin.defs#reportView'


@dataclass
class ReportViewDetail(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        id: Id.
        reasonType: Reason type.
        reason: Reason.
        subject: Subject.
        reportedBy: Reported by.
        createdAt: Created at.
        resolvedByActions: Resolved by actions.
    """

    createdAt: str
    id: int
    reasonType: 'models.ComAtprotoModerationDefs.ReasonType'
    reportedBy: str
    resolvedByActions: List['models.ComAtprotoAdminDefs.ActionView']
    subject: Union[
        'models.ComAtprotoAdminDefs.RepoView',
        'models.ComAtprotoAdminDefs.RepoViewNotFound',
        'models.ComAtprotoAdminDefs.RecordView',
        'models.ComAtprotoAdminDefs.RecordViewNotFound',
        'Dict[str, Any]',
    ]
    reason: Optional[str] = None

    _type: str = 'com.atproto.admin.defs#reportViewDetail'


@dataclass
class RepoView(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        did: Did.
        handle: Handle.
        email: Email.
        relatedRecords: Related records.
        indexedAt: Indexed at.
        moderation: Moderation.
        invitedBy: Invited by.
        invitesDisabled: Invites disabled.
    """

    did: str
    handle: str
    indexedAt: str
    moderation: 'models.ComAtprotoAdminDefs.Moderation'
    relatedRecords: List['base.RecordModelBase']
    email: Optional[str] = None
    invitedBy: Optional['models.ComAtprotoServerDefs.InviteCode'] = None
    invitesDisabled: Optional[bool] = None

    _type: str = 'com.atproto.admin.defs#repoView'


@dataclass
class RepoViewDetail(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        did: Did.
        handle: Handle.
        email: Email.
        relatedRecords: Related records.
        indexedAt: Indexed at.
        moderation: Moderation.
        labels: Labels.
        invitedBy: Invited by.
        invites: Invites.
        invitesDisabled: Invites disabled.
    """

    did: str
    handle: str
    indexedAt: str
    moderation: 'models.ComAtprotoAdminDefs.ModerationDetail'
    relatedRecords: List['base.RecordModelBase']
    email: Optional[str] = None
    invitedBy: Optional['models.ComAtprotoServerDefs.InviteCode'] = None
    invites: Optional[List['models.ComAtprotoServerDefs.InviteCode']] = None
    invitesDisabled: Optional[bool] = None
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None

    _type: str = 'com.atproto.admin.defs#repoViewDetail'


@dataclass
class RepoViewNotFound(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        did: Did.
    """

    did: str

    _type: str = 'com.atproto.admin.defs#repoViewNotFound'


@dataclass
class RepoRef(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        did: Did.
    """

    did: str

    _type: str = 'com.atproto.admin.defs#repoRef'


@dataclass
class RecordView(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        uri: Uri.
        cid: Cid.
        value: Value.
        blobCids: Blob cids.
        indexedAt: Indexed at.
        moderation: Moderation.
        repo: Repo.
    """

    blobCids: List[str]
    cid: str
    indexedAt: str
    moderation: 'models.ComAtprotoAdminDefs.Moderation'
    repo: 'models.ComAtprotoAdminDefs.RepoView'
    uri: str
    value: 'base.RecordModelBase'

    _type: str = 'com.atproto.admin.defs#recordView'


@dataclass
class RecordViewDetail(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        uri: Uri.
        cid: Cid.
        value: Value.
        blobs: Blobs.
        labels: Labels.
        indexedAt: Indexed at.
        moderation: Moderation.
        repo: Repo.
    """

    blobs: List['models.ComAtprotoAdminDefs.BlobView']
    cid: str
    indexedAt: str
    moderation: 'models.ComAtprotoAdminDefs.ModerationDetail'
    repo: 'models.ComAtprotoAdminDefs.RepoView'
    uri: str
    value: 'base.RecordModelBase'
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None

    _type: str = 'com.atproto.admin.defs#recordViewDetail'


@dataclass
class RecordViewNotFound(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        uri: Uri.
    """

    uri: str

    _type: str = 'com.atproto.admin.defs#recordViewNotFound'


@dataclass
class Moderation(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        currentAction: Current action.
    """

    currentAction: Optional['models.ComAtprotoAdminDefs.ActionViewCurrent'] = None

    _type: str = 'com.atproto.admin.defs#moderation'


@dataclass
class ModerationDetail(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        currentAction: Current action.
        actions: Actions.
        reports: Reports.
    """

    actions: List['models.ComAtprotoAdminDefs.ActionView']
    reports: List['models.ComAtprotoAdminDefs.ReportView']
    currentAction: Optional['models.ComAtprotoAdminDefs.ActionViewCurrent'] = None

    _type: str = 'com.atproto.admin.defs#moderationDetail'


@dataclass
class BlobView(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        cid: Cid.
        mimeType: Mime type.
        size: Size.
        createdAt: Created at.
        details: Details.
        moderation: Moderation.
    """

    cid: str
    createdAt: str
    mimeType: str
    size: int
    details: Optional[
        Union['models.ComAtprotoAdminDefs.ImageDetails', 'models.ComAtprotoAdminDefs.VideoDetails', 'Dict[str, Any]']
    ] = None
    moderation: Optional['models.ComAtprotoAdminDefs.Moderation'] = None

    _type: str = 'com.atproto.admin.defs#blobView'


@dataclass
class ImageDetails(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        width: Width.
        height: Height.
    """

    height: int
    width: int

    _type: str = 'com.atproto.admin.defs#imageDetails'


@dataclass
class VideoDetails(base.ModelBase):

    """Definition model for :obj:`com.atproto.admin.defs`.

    Attributes:
        width: Width.
        height: Height.
        length: Length.
    """

    height: int
    length: int
    width: int

    _type: str = 'com.atproto.admin.defs#videoDetails'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Type, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.admin.takeModerationAction`.

    Attributes:
        action: Action.
        subject: Subject.
        subjectBlobCids: Subject blob cids.
        createLabelVals: Create label vals.
        negateLabelVals: Negate label vals.
        reason: Reason.
        createdBy: Created by.
    """

    action: str
    createdBy: str
    reason: str
    subject: Union['models.ComAtprotoAdminDefs.RepoRef', 'models.ComAtprotoRepoStrongRef.Main', 'Dict[str, Any]']
    createLabelVals: Optional[List[str]] = None
    negateLabelVals: Optional[List[str]] = None
    subjectBlobCids: Optional[List[str]] = None


#: Response reference to :obj:`models.ComAtprotoAdminDefs.ActionView` model.
ResponseRef: Type[models.ComAtprotoAdminDefs.ActionView] = models.ComAtprotoAdminDefs.ActionView



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Type

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.admin.getModerationAction`.

    Attributes:
        id: Id.
    """

    id: int


#: Response reference to :obj:`models.ComAtprotoAdminDefs.ActionViewDetail` model.
ResponseRef: Type[models.ComAtprotoAdminDefs.ActionViewDetail] = models.ComAtprotoAdminDefs.ActionViewDetail



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.feed.getLikes`.

    Attributes:
        uri: Uri.
        cid: Cid.
        limit: Limit.
        cursor: Cursor.
    """

    uri: str
    cid: Optional[str] = None
    cursor: Optional[str] = None
    limit: Optional[int] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.feed.getLikes`.

    Attributes:
        uri: Uri.
        cid: Cid.
        cursor: Cursor.
        likes: Likes.
    """

    likes: List['models.AppBskyFeedGetLikes.Like']
    uri: str
    cid: Optional[str] = None
    cursor: Optional[str] = None


@dataclass
class Like(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.getLikes`.

    Attributes:
        indexedAt: Indexed at.
        createdAt: Created at.
        actor: Actor.
    """

    actor: 'models.AppBskyActorDefs.ProfileView'
    createdAt: str
    indexedAt: str

    _type: str = 'app.bsky.feed.getLikes#like'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.sync.listBlobs`.

    Attributes:
        did: The DID of the repo.
        latest: The most recent commit.
        earliest: The earliest commit to start from.
    """

    did: str
    earliest: Optional[str] = None
    latest: Optional[str] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.sync.listBlobs`.

    Attributes:
        cids: Cids.
    """

    cids: List[str]



from atproto.xrpc_client.client.base import ClientBase
from atproto.xrpc_client.namespaces import sync_ns

# TODO(MarshalX): This file should be autogenerated!


class ClientRaw(ClientBase):
    """Group all root namespaces"""

    com: sync_ns.ComNamespace
    bsky: sync_ns.BskyNamespace

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.com = sync_ns.ComNamespace(self)
        self.bsky = sync_ns.BskyNamespace(self)



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass

from atproto.xrpc_client.models import base


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`com.atproto.repo.strongRef`.

    Attributes:
        uri: Uri.
        cid: Cid.
    """

    cid: str
    uri: str

    _type: str = 'com.atproto.repo.strongRef'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client.models import base


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.getSession`.

    Attributes:
        handle: Handle.
        did: Did.
        email: Email.
    """

    did: str
    handle: str
    email: Optional[str] = None



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass

from atproto.xrpc_client.models import base


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.refreshSession`.

    Attributes:
        accessJwt: Access jwt.
        refreshJwt: Refresh jwt.
        handle: Handle.
        did: Did.
    """

    accessJwt: str
    did: str
    handle: str
    refreshJwt: str



import dataclasses
import json
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar, Union

from dacite import Config, exceptions, from_dict

from atproto.cid import CID
from atproto.exceptions import (
    MissingValueError,
    ModelError,
    ModelFieldError,
    UnexpectedFieldError,
    WrongTypeError,
)
from atproto.xrpc_client.models.base import RecordModelBase
from atproto.xrpc_client.models.blob_ref import BlobRef
from atproto.xrpc_client.models.type_conversion import RECORD_TYPE_TO_MODEL_CLASS

if TYPE_CHECKING:
    from atproto.xrpc_client.request import Response

M = TypeVar('M')


def _record_model_type_hook(data: dict) -> RecordModelBase:
    record_type = data.pop('$type')
    return get_or_create_model(data, RECORD_TYPE_TO_MODEL_CLASS[record_type])


_TYPE_HOOKS = {
    BlobRef: lambda ref: BlobRef.from_dict(ref),
    CID: lambda ref: CID.decode(ref),
    RecordModelBase: _record_model_type_hook,
}
_DACITE_CONFIG = Config(type_hooks=_TYPE_HOOKS)


def get_or_create_model(model_data: Union[dict], model: Type[M]) -> Optional[M]:
    if model_data is None:
        return None

    if isinstance(model_data, model):
        return model_data

    try:
        # validate unexpected fields
        model(**model_data)

        return from_dict(model, model_data, config=_DACITE_CONFIG)
    except TypeError as e:
        # FIXME(MarshalX): "Params missing 1 required positional argument: 'rkey'" should raise another error
        msg = str(e).replace('__init__()', model.__name__)
        raise UnexpectedFieldError(msg)
    except exceptions.MissingValueError as e:
        raise MissingValueError(str(e))
    except exceptions.WrongTypeError as e:
        raise WrongTypeError(str(e))
    except exceptions.DaciteFieldError as e:
        raise ModelFieldError(str(e))
    except exceptions.DaciteError as e:
        raise ModelError(str(e))


def get_response_model(response: 'Response', model: Type[M]) -> Optional[M]:
    if model is bool:
        # Could not be False? Because the exception with errors will be raised from the server
        return response.success

    return get_or_create_model(response.content, model)


def _handle_dict_key(key: str) -> str:
    if key == '_type':  # System field. Replaced to original $ symbol because is not allowed in Python.
        return '$type'

    return key


def _handle_dict_value(ref: Any) -> Any:
    if isinstance(ref, BlobRef):
        return ref.to_dict()
    elif isinstance(ref, CID):
        return ref.encode()

    return ref


def _model_as_dict_factory(value):
    # exclude None values and process keys and values
    return {_handle_dict_key(k): _handle_dict_value(v) for k, v in value if v is not None}


def get_model_as_dict(model) -> dict:
    if model == BlobRef:
        return model.to_dict()

    if not dataclasses.is_dataclass(model):
        raise ModelError('Invalid model')

    return dataclasses.asdict(model, dict_factory=_model_as_dict_factory)


def get_model_as_json(model) -> str:
    return json.dumps(get_model_as_dict(model))


def is_json(json_data: Union[str, bytes]):
    if isinstance(json_data, bytes):
        json_data.decode('UTF-8')

    try:
        json.loads(json_data)
        return True
    except Exception:  # noqa
        return False



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`com.atproto.repo.getRecord`.

    Attributes:
        repo: The handle or DID of the repo.
        collection: The NSID of the record collection.
        rkey: The key of the record.
        cid: The CID of the version of the record. If not specified, then return the most recent version.
    """

    collection: str
    repo: str
    rkey: str
    cid: Optional[str] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.repo.getRecord`.

    Attributes:
        uri: Uri.
        cid: Cid.
        value: Value.
    """

    uri: str
    value: 'base.RecordModelBase'
    cid: Optional[str] = None



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Any, Dict, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.recordWithMedia`.

    Attributes:
        record: Record.
        media: Media.
    """

    media: Union['models.AppBskyEmbedImages.Main', 'models.AppBskyEmbedExternal.Main', 'Dict[str, Any]']
    record: 'models.AppBskyEmbedRecord.Main'

    _type: str = 'app.bsky.embed.recordWithMedia'


@dataclass
class View(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.recordWithMedia`.

    Attributes:
        record: Record.
        media: Media.
    """

    media: Union['models.AppBskyEmbedImages.View', 'models.AppBskyEmbedExternal.View', 'Dict[str, Any]']
    record: 'models.AppBskyEmbedRecord.View'

    _type: str = 'app.bsky.embed.recordWithMedia#view'



from .car import CAR
from .cid import CID
from .nsid import NSID
from .uri import AtUri
from .xrpc_client import models
from .xrpc_client.client.async_client import AsyncClient
from .xrpc_client.client.client import Client

__all__ = [
    'AsyncClient',
    'Client',
    'models',
    'NSID',
    'CAR',
    'CID',
    'AtUri',
]



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base
from atproto.xrpc_client.models.blob_ref import BlobRef


@dataclass
class Main(base.RecordModelBase):

    """Record model for :obj:`app.bsky.graph.list`.

    Attributes:
        purpose: Purpose.
        name: Name.
        description: Description.
        descriptionFacets: Description facets.
        avatar: Avatar.
        createdAt: Created at.
    """

    createdAt: str
    name: str
    purpose: 'models.AppBskyGraphDefs.ListPurpose'
    avatar: Optional[BlobRef] = None
    description: Optional[str] = None
    descriptionFacets: Optional[List['models.AppBskyRichtextFacet.Main']] = None

    _type: str = 'app.bsky.graph.list'



from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Dict, Union

from atproto.xrpc_client.client.auth import get_jwt_payload

if TYPE_CHECKING:
    from atproto.xrpc_client import models


class SessionMethodsMixin:
    def _should_refresh_session(self) -> bool:
        expired_at = datetime.fromtimestamp(self._access_jwt_payload.exp, tz=timezone.utc)
        expired_at = expired_at - timedelta(minutes=15)  # let's update token a bit later than required

        datetime_now = datetime.now(timezone.utc)

        return datetime_now > expired_at

    def _set_session(
        self,
        session: Union[
            'models.ComAtprotoServerCreateSession.Response', 'models.ComAtprotoServerRefreshSession.Response'
        ],
    ) -> None:
        self._access_jwt = session.accessJwt
        self._access_jwt_payload = get_jwt_payload(session.accessJwt)

        self._refresh_jwt = session.refreshJwt
        self._refresh_jwt_payload = get_jwt_payload(session.refreshJwt)

        self._set_auth_headers(session.accessJwt)

    @staticmethod
    def _get_auth_headers(token: str) -> Dict[str, str]:
        return {'Authorization': f'Bearer {token}'}

    def _set_auth_headers(self, token: str) -> None:
        self.request.set_additional_headers(self._get_auth_headers(token))  # noqa



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.client.async_raw import AsyncClientRaw
from atproto.xrpc_client.client.methods_mixin import SessionMethodsMixin

if TYPE_CHECKING:
    from atproto.xrpc_client.client.auth import JwtPayload
    from atproto.xrpc_client.client.base import InvokeType
    from atproto.xrpc_client.request import Response


class AsyncClient(AsyncClientRaw, SessionMethodsMixin):
    """High-level client for XRPC of ATProto."""

    def __init__(self, base_url: str = None):
        super().__init__(base_url)

        self._access_jwt: Optional[str] = None
        self._access_jwt_payload: Optional['JwtPayload'] = None

        self._refresh_jwt: Optional[str] = None
        self._refresh_jwt_payload: Optional['JwtPayload'] = None

        self.me: Optional[models.AppBskyActorDefs.ProfileViewDetailed] = None

    async def _invoke(self, invoke_type: 'InvokeType', **kwargs) -> 'Response':
        if self.me and self._should_refresh_session():
            await self._refresh_and_set_session()

        return await super()._invoke(invoke_type, **kwargs)

    async def _get_and_set_session(self, login: str, password: str) -> models.ComAtprotoServerCreateSession.Response:
        session = await self.com.atproto.server.create_session(
            models.ComAtprotoServerCreateSession.Data(login, password)
        )
        self._set_session(session)

        return session

    async def _refresh_and_set_session(self) -> models.ComAtprotoServerRefreshSession.Response:
        refresh_session = await self.com.atproto.server.refresh_session(
            headers=self._get_auth_headers(self._refresh_jwt)
        )
        self._set_session(refresh_session)

        return refresh_session

    async def login(self, login: str, password: str) -> models.AppBskyActorGetProfile.ResponseRef:
        """Authorize client and get profile info.

        Args:
            login: Handle/username of the account.
            password: Password of the account. Could be app specific one.

        Returns:
            :obj:`models.AppBskyActorGetProfile.ResponseRef`: Profile information.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """

        session = await self._get_and_set_session(login, password)
        self.me = await self.bsky.actor.get_profile(models.AppBskyActorGetProfile.Params(session.handle))

        return self.me

    async def send_post(
        self,
        text: str,
        profile_identify: Optional[str] = None,
        reply_to: Optional[Union[models.AppBskyFeedPost.ReplyRef, models.AppBskyFeedDefs.ReplyRef]] = None,
        embed: Optional[
            Union[
                'models.AppBskyEmbedImages.Main',
                'models.AppBskyEmbedExternal.Main',
                'models.AppBskyEmbedRecord.Main',
                'models.AppBskyEmbedRecordWithMedia.Main',
            ]
        ] = None,
    ) -> models.ComAtprotoRepoCreateRecord.Response:
        """Send post.

        Note:
            If `profile_identify` is not provided will be sent to the current profile.

        Args:
            text: Text of the post.
            profile_identify: Handle or DID. Where to send post.
            reply_to: Root and parent of the post to reply to.
            embed: Embed models that should be attached to the post.

        Returns:
            :obj:`models.ComAtprotoRepoCreateRecord.Response`: Reference to the created post record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """

        repo = self.me.did
        if profile_identify:
            repo = profile_identify

        return await self.com.atproto.repo.create_record(
            models.ComAtprotoRepoCreateRecord.Data(
                repo=repo,
                collection='app.bsky.feed.post',
                record=models.AppBskyFeedPost.Main(
                    createdAt=datetime.now().isoformat(), text=text, reply=reply_to, embed=embed
                ),
            )
        )

    async def send_image(
        self,
        text: str,
        image: bytes,
        image_alt: str,
        profile_identify: Optional[str] = None,
        reply_to: Optional[Union[models.AppBskyFeedPost.ReplyRef, models.AppBskyFeedDefs.ReplyRef]] = None,
    ) -> models.ComAtprotoRepoCreateRecord.Response:
        """Send post with attached image.

        Note:
            If `profile_identify` is not provided will be sent to the current profile.

        Args:
            text: Text of the post.
            image: Binary image to attach.
            image_alt: Text version of the image
            profile_identify: Handle or DID. Where to send post.
            reply_to: Root and parent of the post to reply to.

        Returns:
            :obj:`models.ComAtprotoRepoCreateRecord.Response`: Reference to the created post record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """

        upload = await self.com.atproto.repo.upload_blob(image)
        images = [models.AppBskyEmbedImages.Image(alt=image_alt, image=upload.blob)]
        return await self.send_post(
            text,
            profile_identify=profile_identify,
            reply_to=reply_to,
            embed=models.AppBskyEmbedImages.Main(images=images),
        )

    async def like(self, subject: models.ComAtprotoRepoStrongRef.Main) -> models.ComAtprotoRepoCreateRecord.Response:
        """Like the post.

        Args:
            subject: Reference to the post that should be liked.

        Returns:
            :obj:`models.ComAtprotoRepoCreateRecord.Response`: Reference to the created like record.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """

        return await self.com.atproto.repo.create_record(
            models.ComAtprotoRepoCreateRecord.Data(
                repo=self.me.did,
                collection='app.bsky.feed.like',
                record=models.AppBskyFeedLike.Main(createdAt=datetime.now().isoformat(), subject=subject),
            )
        )

    async def unlike(self, record_key: str, profile_identify: Optional[str] = None) -> bool:
        """Unlike the post.

        Args:
            record_key: ID (slog) of the post.
            profile_identify: Handler or DID. Who did the like.

        Returns:
            :obj:`bool`: Success status.

        Raises:
            :class:`atproto.exceptions.AtProtocolError`: Base exception.
        """

        repo = self.me.did
        if profile_identify:
            repo = profile_identify

        return await self.com.atproto.repo.delete_record(
            models.ComAtprotoRepoDeleteRecord.Data(
                collection='app.bsky.feed.like',
                repo=repo,
                rkey=record_key,
            )
        )



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Params(base.ParamsModelBase):

    """Parameters model for :obj:`app.bsky.feed.getPostThread`.

    Attributes:
        uri: Uri.
        depth: Depth.
    """

    uri: str
    depth: Optional[int] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`app.bsky.feed.getPostThread`.

    Attributes:
        thread: Thread.
    """

    thread: Union[
        'models.AppBskyFeedDefs.ThreadViewPost',
        'models.AppBskyFeedDefs.NotFoundPost',
        'models.AppBskyFeedDefs.BlockedPost',
        'Dict[str, Any]',
    ]



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class ReplyRef(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.post`.

    Attributes:
        root: Root.
        parent: Parent.
    """

    parent: 'models.ComAtprotoRepoStrongRef.Main'
    root: 'models.ComAtprotoRepoStrongRef.Main'

    _type: str = 'app.bsky.feed.post#replyRef'


@dataclass
class Entity(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.post`. Deprecated: use facets instead.

    Attributes:
        index: Index.
        type: Expected values are 'mention' and 'link'.
        value: Value.
    """

    index: 'models.AppBskyFeedPost.TextSlice'
    type: str
    value: str

    _type: str = 'app.bsky.feed.post#entity'


@dataclass
class TextSlice(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.post`. Deprecated. Use app.bsky.richtext instead -- A text segment. Start is inclusive, end is exclusive. Indices are for utf16-encoded strings.

    Attributes:
        start: Start.
        end: End.
    """

    end: int
    start: int

    _type: str = 'app.bsky.feed.post#textSlice'


@dataclass
class Main(base.RecordModelBase):

    """Record model for :obj:`app.bsky.feed.post`.

    Attributes:
        text: Text.
        entities: Deprecated: replaced by app.bsky.richtext.facet.
        facets: Facets.
        reply: Reply.
        embed: Embed.
        createdAt: Created at.
    """

    createdAt: str
    text: str
    embed: Optional[
        Union[
            'models.AppBskyEmbedImages.Main',
            'models.AppBskyEmbedExternal.Main',
            'models.AppBskyEmbedRecord.Main',
            'models.AppBskyEmbedRecordWithMedia.Main',
            'Dict[str, Any]',
        ]
    ] = None
    entities: Optional[List['models.AppBskyFeedPost.Entity']] = None
    facets: Optional[List['models.AppBskyRichtextFacet.Main']] = None
    reply: Optional['models.AppBskyFeedPost.ReplyRef'] = None

    _type: str = 'app.bsky.feed.post'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.server.createSession`.

    Attributes:
        identifier: Handle or other identifier supported by the server for the authenticating user.
        password: Password.
    """

    identifier: str
    password: str


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.createSession`.

    Attributes:
        accessJwt: Access jwt.
        refreshJwt: Refresh jwt.
        handle: Handle.
        did: Did.
        email: Email.
    """

    accessJwt: str
    did: str
    handle: str
    refreshJwt: str
    email: Optional[str] = None



from atproto.xrpc_client.client.base import AsyncClientBase
from atproto.xrpc_client.namespaces import async_ns

# TODO(MarshalX): This file should be autogenerated!


class AsyncClientRaw(AsyncClientBase):
    """Group all root namespaces"""

    com: async_ns.ComNamespace
    bsky: async_ns.BskyNamespace

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.com = async_ns.ComNamespace(self)
        self.bsky = async_ns.BskyNamespace(self)



from typing import Dict, List, Union

from atproto.lexicon import models
from atproto.lexicon.parser import lexicon_parse_dir
from atproto.nsid import NSID

_LEX_DEF_TYPES_FOR_PARAMS = {models.LexDefinitionType.QUERY, models.LexDefinitionType.PROCEDURE}
_LEX_DEF_TYPES_FOR_RESPONSES = {models.LexDefinitionType.QUERY, models.LexDefinitionType.PROCEDURE}
_LEX_DEF_TYPES_FOR_REFS = {models.LexDefinitionType.QUERY, models.LexDefinitionType.PROCEDURE}
_LEX_DEF_TYPES_FOR_DATA = {models.LexDefinitionType.PROCEDURE}
_LEX_DEF_TYPES_FOR_RECORDS = {models.LexDefinitionType.RECORD}
_LEX_DEF_TYPES_FOR_DEF = {
    models.LexDefinitionType.OBJECT,
    models.LexPrimitiveType.STRING,
    models.LexDefinitionType.TOKEN,
    models.LexDefinitionType.ARRAY,
}

LexDefs = Dict[
    str,
    Union[
        models.LexXrpcProcedure,
        models.LexXrpcQuery,
        models.LexObject,
        models.LexToken,
        models.LexString,
        models.LexRecord,
    ],
]
LexDB = Dict[NSID, LexDefs]


def _filter_defs_by_type(defs: Dict[str, models.LexDefinition], def_types: set) -> LexDefs:
    return {k: v for k, v in defs.items() if v.type in def_types}


def _build_nsid_to_defs_map(lexicons: List[models.LexiconDoc], def_types: set) -> LexDB:
    result = {}

    for lexicon in lexicons:
        nsid = NSID.from_str(lexicon.id)
        defs = _filter_defs_by_type(lexicon.defs, def_types)
        if defs:
            result[nsid] = defs

    return result


def build_params_models() -> LexDB:
    return _build_nsid_to_defs_map(lexicon_parse_dir(), _LEX_DEF_TYPES_FOR_PARAMS)


def build_data_models() -> LexDB:
    return _build_nsid_to_defs_map(lexicon_parse_dir(), _LEX_DEF_TYPES_FOR_DATA)


def build_response_models() -> LexDB:
    return _build_nsid_to_defs_map(lexicon_parse_dir(), _LEX_DEF_TYPES_FOR_RESPONSES)


def build_def_models() -> LexDB:
    return _build_nsid_to_defs_map(lexicon_parse_dir(), _LEX_DEF_TYPES_FOR_DEF)


def build_record_models() -> LexDB:
    return _build_nsid_to_defs_map(lexicon_parse_dir(), _LEX_DEF_TYPES_FOR_RECORDS)


def build_refs_models() -> LexDB:
    return _build_nsid_to_defs_map(lexicon_parse_dir(), _LEX_DEF_TYPES_FOR_REFS)


if __name__ == '__main__':
    build_params_models()
    build_data_models()
    build_response_models()
    build_def_models()
    build_record_models()
    build_refs_models()



from typing import Dict, List, NamedTuple, Union

from atproto.lexicon.models import (
    LexDefinition,
    LexDefinitionType,
    LexiconDoc,
    LexRecord,
    LexXrpcProcedure,
    LexXrpcQuery,
)
from atproto.lexicon.parser import lexicon_parse_dir
from atproto.nsid import NSID

_VALID_LEX_DEF_TYPES = {LexDefinitionType.QUERY, LexDefinitionType.PROCEDURE, LexDefinitionType.RECORD}
_VALID_LEX_DEF_TYPE = Union[LexXrpcProcedure, LexXrpcQuery, LexRecord]


def _filter_namespace_valid_definitions(definitions: Dict[str, LexDefinition]) -> Dict[str, LexDefinition]:
    result = {}

    for def_name, def_content in definitions.items():
        if def_content.type in _VALID_LEX_DEF_TYPES:
            result[def_name] = def_content

    return result


def get_definition_by_name(name: str, defs: Dict[str, LexDefinition]) -> LexDefinition:
    if name in defs:
        return defs[name]

    return defs['main']


class ObjectInfo(NamedTuple):
    name: str
    nsid: NSID
    definition: _VALID_LEX_DEF_TYPE


class MethodInfo(ObjectInfo):
    pass


class RecordInfo(ObjectInfo):
    pass


def _enrich_namespace_tree(root: dict, nsid: NSID, defs: Dict[str, LexDefinition]) -> None:
    segments_count = len(nsid.segments)
    for path_level, segment in enumerate(nsid.segments):
        # if method
        if path_level == segments_count - 1:
            definition = get_definition_by_name(segment, defs)

            model_class = MethodInfo
            if definition.type is LexDefinitionType.RECORD:
                model_class = RecordInfo

            # TODO(MarshalX): fake records as namespaces with methods to be able to reuse code of generator?

            root.append(model_class(name=segment, nsid=nsid, definition=definition))
            continue

        if segment not in root:
            # if end of method's path
            if path_level == segments_count - 2:
                root[segment] = []
            else:
                root[segment] = {}
        root = root[segment]


def build_namespace_tree(lexicons: List[LexiconDoc]) -> dict:
    namespace_tree = {}
    for lexicon in lexicons:
        defs = _filter_namespace_valid_definitions(lexicon.defs)
        if not defs:
            continue

        nsid = NSID.from_str(lexicon.id)
        _enrich_namespace_tree(namespace_tree, nsid, defs)

    return namespace_tree


def build_namespaces() -> dict:
    lexicons = lexicon_parse_dir()
    namespace_tree = build_namespace_tree(lexicons)

    return namespace_tree


if __name__ == '__main__':
    build_namespaces()



import re
import subprocess
from pathlib import Path
from typing import List, Tuple

from atproto.nsid import NSID

DISCLAIMER = [
    '# THIS IS THE AUTO-GENERATED CODE. DON\'T EDIT IT BY HANDS!',
    '# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.',
    '# This file is part of Python atproto SDK. Licenced under MIT.',
]
_MAX_DISCLAIMER_LEN = max([len(s) for s in DISCLAIMER])
DISCLAIMER = '\n'.join(DISCLAIMER)
DISCLAIMER = f'{"#" * _MAX_DISCLAIMER_LEN}\n{DISCLAIMER}\n{"#" * _MAX_DISCLAIMER_LEN}\n\n'

PARAMS_MODEL = 'Params'
INPUT_MODEL = 'Data'
OUTPUT_MODEL = 'Response'


def format_code(filepath: Path) -> None:
    subprocess.run(['black', '--quiet', filepath])
    subprocess.run(['isort', '--quiet', filepath])


def append_code(filepath: Path, code: str) -> None:
    _write_code(filepath, code, append=True)


def write_code(filepath: Path, code: str) -> None:
    _write_code(filepath, code)


def _write_code(filepath: Path, code: str, *, append: bool = False) -> None:
    filepath.parent.mkdir(exist_ok=True, parents=True)

    mode = 'w'
    if append:
        mode = 'a'

    with open(filepath, mode=mode, encoding='UTF-8') as f:
        f.write(code)


def get_file_path_parts(nsid: NSID) -> List[str]:
    return nsid.segments[:-1] + [f'{convert_camel_case_to_snake_case(nsid.name)}.py']


def get_import_path_old(nsid: NSID) -> str:
    return '.'.join(nsid.segments[:-1] + [f'{convert_camel_case_to_snake_case(nsid.name)}'])


def get_import_path(nsid: NSID) -> str:
    nsid_parts = nsid.segments[:-1] + camel_case_split(nsid.name)
    alias_name = ''.join([p.capitalize() for p in nsid_parts])
    return alias_name


def convert_camel_case_to_snake_case(string: str) -> str:
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()


def camel_case_split(string: str):
    # regex by chatgpt
    return re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)', string)


def gen_description_by_camel_case_name(name: str):
    words = camel_case_split(name)
    words = [w.lower() for w in words]
    words[0] = words[0].capitalize()
    return ' '.join(words)


def sort_dict_by_key(d: dict) -> dict:
    return dict(sorted(d.items()))


def get_code_intent(level: int):
    return ' ' * 4 * level


def join_code(lines: List[str]) -> str:
    return '\n'.join(lines)


def get_sync_async_keywords(*, sync: bool) -> Tuple[str, str]:
    definition, call = 'async ', 'await '
    if sync:
        definition, call = '', ''

    return definition, call


def capitalize_first_symbol(string: str) -> str:
    if string and string[0].islower():
        chars = [c for c in string[1:]]
        chars.insert(0, string[0].upper())
        string = ''.join(chars)

    return string



from pathlib import Path
from typing import List, Set, Union

from atproto.codegen import (
    DISCLAIMER,
    INPUT_MODEL,
    OUTPUT_MODEL,
    PARAMS_MODEL,
    convert_camel_case_to_snake_case,
    format_code,
    gen_description_by_camel_case_name,
)
from atproto.codegen import get_code_intent as _
from atproto.codegen import (
    get_import_path,
    get_sync_async_keywords,
    join_code,
    sort_dict_by_key,
    write_code,
)
from atproto.codegen.namespaces.builder import MethodInfo, RecordInfo, build_namespaces
from atproto.lexicon.models import (
    LexDefinitionType,
    LexObject,
    LexRef,
    LexXrpcProcedure,
)
from atproto.nsid import NSID

_NAMESPACES_OUTPUT_DIR = Path(__file__).parent.parent.parent.joinpath('xrpc_client', 'namespaces')
_NAMESPACES_CLIENT_FILE_PATH = _NAMESPACES_OUTPUT_DIR.joinpath('client', 'raw.py')

_NAMESPACES_SYNC_FILENAME = 'sync_ns.py'
_NAMESPACES_ASYNC_FILENAME = 'async_ns.py'

_NAMESPACE_SUFFIX = 'Namespace'
_RECORD_SUFFIX = 'RecordNamespace'


def get_namespace_name(path_part: str) -> str:
    return f'{path_part.capitalize()}{_NAMESPACE_SUFFIX}'


def get_record_name(path_part: str) -> str:
    return f'{path_part.capitalize()}{_RECORD_SUFFIX}'


def _get_namespace_imports() -> str:
    lines = [
        DISCLAIMER,
        'from dataclasses import dataclass, field',
        'from typing import Optional, Union',
        '',
        'from atproto.xrpc_client import models',
        'from atproto.xrpc_client.models.utils import get_or_create_model, get_response_model',
        'from atproto.xrpc_client.namespaces.base import DefaultNamespace, NamespaceBase',
    ]

    return join_code(lines)


def _get_namespace_class_def(name: str) -> str:
    lines = ['@dataclass', f'class {get_namespace_name(name)}(NamespaceBase):']

    return join_code(lines)


def _get_sub_namespaces_block(sub_namespaces: dict) -> str:
    lines = []

    sub_namespaces = sort_dict_by_key(sub_namespaces)
    for sub_namespace in sub_namespaces.keys():
        lines.append(
            f"{_(1)}{sub_namespace}: '{get_namespace_name(sub_namespace)}' = field(default_factory=DefaultNamespace)"
        )

    return join_code(lines)


def _get_post_init_method(sub_namespaces: dict) -> str:
    lines = [f'{_(1)}def __post_init__(self):']

    sub_namespaces = sort_dict_by_key(sub_namespaces)
    for sub_namespace in sub_namespaces.keys():
        lines.append(f'{_(2)}self.{sub_namespace} = {get_namespace_name(sub_namespace)}(self._client)')

    # TODO(MarshalX): add support for records

    return join_code(lines)


def _get_method_docstring(method_info: MethodInfo) -> str:
    method_desc = f'{gen_description_by_camel_case_name(method_info.name)}.'
    if method_info.definition.description:
        method_desc = method_info.definition.description
    if method_desc[-1] not in {'.', '?', '!'}:
        method_desc += '.'

    doc_string = [f'{_(2)}"""{method_desc}', '', f'{_(2)}Args:']

    presented_args = _get_namespace_method_signature_args_names(method_info)
    if 'params' in presented_args:
        doc_string.append(f'{_(3)}params: Parameters.')
    if 'data_schema' in presented_args:
        doc_string.append(f'{_(3)}data: Input data.')
    if 'data_alias' in presented_args:
        doc_string.append(f'{_(3)}data: Input data alias.')

    doc_string.append(f'{_(3)}**kwargs: Arbitrary arguments to HTTP request.')
    doc_string.append('')

    doc_string.append(f'{_(2)}Returns:')

    return_type = _get_namespace_method_return_type(method_info)

    return_type_desc = 'Output model'
    if return_type == 'bool':
        return_type_desc = 'Success status'

    doc_string.append(f'{_(3)}:obj:`{return_type}`: {return_type_desc}.')
    doc_string.append('')

    doc_string.append(f'{_(2)}Raises:')
    doc_string.append(f'{_(3)}:class:`atproto.exceptions.AtProtocolError`: Base exception.')

    doc_string.append(f'{_(2)}"""')
    doc_string.append('')

    return join_code(doc_string)


def _get_namespace_method_body(method_info: MethodInfo, *, sync: bool) -> str:
    d, c = get_sync_async_keywords(sync=sync)

    lines = [_get_method_docstring(method_info)]

    presented_args = _get_namespace_method_signature_args_names(method_info)
    presented_args.remove('self')

    def _override_arg_line(name: str, model_name: str) -> str:
        model_path = f'models.{get_import_path(method_info.nsid)}.{model_name}'
        return f'{_(2)}{name} = get_or_create_model({name}, {model_path})'

    invoke_args = [f"'{method_info.nsid}'"]

    if 'params' in presented_args:
        invoke_args.append('params=params')
        lines.append(_override_arg_line('params', PARAMS_MODEL))
    if 'data_schema' in presented_args:
        invoke_args.append('data=data')
        lines.append(_override_arg_line('data', INPUT_MODEL))
    if 'data_alias' in presented_args:
        invoke_args.append('data=data')
    if 'input_encoding' in presented_args:
        input_encoding = method_info.definition.input.encoding
        invoke_args.append(f"input_encoding='{input_encoding}'")
    if 'output_encoding' in presented_args:
        output_encoding = method_info.definition.output.encoding
        invoke_args.append(f"output_encoding='{output_encoding}'")

    invoke_args.append('**kwargs')
    invoke_args_str = ', '.join(invoke_args)

    method_name = 'invoke_query'
    if isinstance(method_info.definition, LexXrpcProcedure):
        method_name = 'invoke_procedure'

    lines.append(f"{_(2)}response = {c}self._client.{method_name}({invoke_args_str})")

    return_type = _get_namespace_method_return_type(method_info)
    lines.append(f"{_(2)}return get_response_model(response, {return_type})")

    return join_code(lines)


def _get_namespace_method_signature_arg(
    name: str, nsid: NSID, model_name: str, *, optional: bool, alias: bool = False
) -> str:
    if alias:
        return f"{name}: 'models.{get_import_path(nsid)}.{model_name}'"

    default_value = ''
    type_hint = f"Union[dict, 'models.{get_import_path(nsid)}.{model_name}']"
    if optional:
        type_hint = f'Optional[{type_hint}]'
        default_value = ' = None'

    return f'{name}: {type_hint}{default_value}'


def _get_namespace_method_signature_args_names(method_info: MethodInfo) -> Set[str]:
    args = {'self'}
    if method_info.definition.parameters:
        args.add('params')

    if method_info.definition.type is LexDefinitionType.PROCEDURE and method_info.definition.input:
        if method_info.definition.input.schema:
            args.add('data_schema')
        else:
            args.add('data_alias')

        if method_info.definition.input.encoding:
            args.add('input_encoding')

    if method_info.definition.output and method_info.definition.output.encoding:
        args.add('output_encoding')

    return args


def _get_namespace_method_signature_args(method_info: MethodInfo) -> str:
    args = ['self']
    optional_args = []

    def _add_arg(arg_def: str, *, optional: bool) -> None:
        if optional:
            optional_args.append(arg_def)
        else:
            args.append(arg_def)

    def is_optional_arg(lex_obj) -> bool:
        return lex_obj.required is None or len(lex_obj.required) == 0

    if method_info.definition.parameters:
        params = method_info.definition.parameters
        is_optional = is_optional_arg(params)

        arg = _get_namespace_method_signature_arg('params', method_info.nsid, PARAMS_MODEL, optional=is_optional)
        _add_arg(arg, optional=is_optional)

    if method_info.definition.type is LexDefinitionType.PROCEDURE and method_info.definition.input:
        schema = method_info.definition.input.schema
        if schema:
            is_optional = is_optional_arg(schema)

            if schema and isinstance(schema, LexObject):
                arg = _get_namespace_method_signature_arg('data', method_info.nsid, INPUT_MODEL, optional=is_optional)
                _add_arg(arg, optional=is_optional)
            else:
                raise ValueError(f'Bad type {type(schema)}')  # probably LexRefVariant
        else:
            arg = _get_namespace_method_signature_arg('data', method_info.nsid, INPUT_MODEL, optional=False, alias=True)
            _add_arg(arg, optional=False)

    args.extend(optional_args)
    args.append('**kwargs')
    return ', '.join(args)


def _get_namespace_method_return_type(method_info: MethodInfo) -> str:
    model_name_suffix = ''
    if method_info.definition.output and isinstance(method_info.definition.output.schema, LexRef):
        # fix collisions with type aliases
        # example of collisions: com.atproto.admin.getRepo, com.atproto.sync.getRepo
        # could be solved by separating models into different folders using segments of NSID
        model_name_suffix = 'Ref'

    return_type = 'bool'  # return success of response
    if method_info.definition.output:
        # example of methods without response: app.bsky.graph.muteActor, app.bsky.graph.muteActor
        return_type = f'models.{get_import_path(method_info.nsid)}.{OUTPUT_MODEL}{model_name_suffix}'

    return return_type


def _get_namespace_method_signature(method_info: MethodInfo, *, sync: bool) -> str:
    d, c = get_sync_async_keywords(sync=sync)

    name = convert_camel_case_to_snake_case(method_info.name)
    args = _get_namespace_method_signature_args(method_info)
    return_type = _get_namespace_method_return_type(method_info)

    return f'{_(1)}{d}def {name}({args}) -> {return_type}:'


def _get_namespace_methods_block(methods_info: List[MethodInfo], sync: bool) -> str:
    lines = []

    methods_info.sort(key=lambda e: e.name)
    for method_info in methods_info:
        lines.append(_get_namespace_method_signature(method_info, sync=sync))
        lines.append(_get_namespace_method_body(method_info, sync=sync))

    return join_code(lines)


def _get_namespace_records_block(records_info: List[RecordInfo]) -> str:
    lines = []

    records_info.sort(key=lambda e: e.name)
    for record_info in records_info:
        lines.append(f"{_(1)}{record_info.name}: '{get_record_name(record_info.name)}' = DefaultNamespace()")

    return join_code(lines)


def _generate_namespace_in_output(namespace_tree: Union[dict, list], output: List[str], *, sync: bool) -> None:
    for node_name, sub_node in namespace_tree.items():
        if isinstance(sub_node, dict):
            output.append(_get_namespace_class_def(node_name))
            output.append(_get_sub_namespaces_block(sub_node))
            output.append(_get_post_init_method(sub_node))

            _generate_namespace_in_output(sub_node, output, sync=sync)

        if isinstance(sub_node, list):
            output.append(_get_namespace_class_def(node_name))

            # TODO(MarshalX): implement later
            # records = [info for info in sub_node if isinstance(info, RecordInfo)]
            # output.append(_get_namespace_records_block(records))
            # TODO(MarshalX): generate namespace record classes!

            methods = [info for info in sub_node if isinstance(info, MethodInfo)]
            output.append(_get_namespace_methods_block(methods, sync=sync))


def generate_namespaces() -> None:
    namespace_tree = build_namespaces()

    for sync in (True, False):
        generated_code_lines_buffer = []
        _generate_namespace_in_output(namespace_tree, generated_code_lines_buffer, sync=sync)

        code = join_code([_get_namespace_imports(), *generated_code_lines_buffer])

        filename = _NAMESPACES_SYNC_FILENAME if sync else _NAMESPACES_ASYNC_FILENAME
        filepath = _NAMESPACES_OUTPUT_DIR.joinpath(filename)

        write_code(filepath, code)

    # TODO(MarshalX): generate ClientRaw as root of namespaces

    format_code(_NAMESPACES_OUTPUT_DIR)
    print('Done')


if __name__ == '__main__':
    generate_namespaces()



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base
from atproto.xrpc_client.models.blob_ref import BlobRef


@dataclass
class Main(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`.

    Attributes:
        external: External.
    """

    external: 'models.AppBskyEmbedExternal.External'

    _type: str = 'app.bsky.embed.external'


@dataclass
class External(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`.

    Attributes:
        uri: Uri.
        title: Title.
        description: Description.
        thumb: Thumb.
    """

    description: str
    title: str
    uri: str
    thumb: Optional[BlobRef] = None

    _type: str = 'app.bsky.embed.external#external'


@dataclass
class View(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`.

    Attributes:
        external: External.
    """

    external: 'models.AppBskyEmbedExternal.ViewExternal'

    _type: str = 'app.bsky.embed.external#view'


@dataclass
class ViewExternal(base.ModelBase):

    """Definition model for :obj:`app.bsky.embed.external`.

    Attributes:
        uri: Uri.
        title: Title.
        description: Description.
        thumb: Thumb.
    """

    description: str
    title: str
    uri: str
    thumb: Optional[str] = None

    _type: str = 'app.bsky.embed.external#viewExternal'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass

from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.admin.updateAccountHandle`.

    Attributes:
        did: Did.
        handle: Handle.
    """

    did: str
    handle: str



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class ProfileViewBasic(base.ModelBase):

    """Definition model for :obj:`app.bsky.actor.defs`.

    Attributes:
        did: Did.
        handle: Handle.
        displayName: Display name.
        avatar: Avatar.
        viewer: Viewer.
        labels: Labels.
    """

    did: str
    handle: str
    avatar: Optional[str] = None
    displayName: Optional[str] = None
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None
    viewer: Optional['models.AppBskyActorDefs.ViewerState'] = None

    _type: str = 'app.bsky.actor.defs#profileViewBasic'


@dataclass
class ProfileView(base.ModelBase):

    """Definition model for :obj:`app.bsky.actor.defs`.

    Attributes:
        did: Did.
        handle: Handle.
        displayName: Display name.
        description: Description.
        avatar: Avatar.
        indexedAt: Indexed at.
        viewer: Viewer.
        labels: Labels.
    """

    did: str
    handle: str
    avatar: Optional[str] = None
    description: Optional[str] = None
    displayName: Optional[str] = None
    indexedAt: Optional[str] = None
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None
    viewer: Optional['models.AppBskyActorDefs.ViewerState'] = None

    _type: str = 'app.bsky.actor.defs#profileView'


@dataclass
class ProfileViewDetailed(base.ModelBase):

    """Definition model for :obj:`app.bsky.actor.defs`.

    Attributes:
        did: Did.
        handle: Handle.
        displayName: Display name.
        description: Description.
        avatar: Avatar.
        banner: Banner.
        followersCount: Followers count.
        followsCount: Follows count.
        postsCount: Posts count.
        indexedAt: Indexed at.
        viewer: Viewer.
        labels: Labels.
    """

    did: str
    handle: str
    avatar: Optional[str] = None
    banner: Optional[str] = None
    description: Optional[str] = None
    displayName: Optional[str] = None
    followersCount: Optional[int] = None
    followsCount: Optional[int] = None
    indexedAt: Optional[str] = None
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None
    postsCount: Optional[int] = None
    viewer: Optional['models.AppBskyActorDefs.ViewerState'] = None

    _type: str = 'app.bsky.actor.defs#profileViewDetailed'


@dataclass
class ViewerState(base.ModelBase):

    """Definition model for :obj:`app.bsky.actor.defs`.

    Attributes:
        muted: Muted.
        mutedByList: Muted by list.
        blockedBy: Blocked by.
        blocking: Blocking.
        following: Following.
        followedBy: Followed by.
    """

    blockedBy: Optional[bool] = None
    blocking: Optional[str] = None
    followedBy: Optional[str] = None
    following: Optional[str] = None
    muted: Optional[bool] = None
    mutedByList: Optional['models.AppBskyGraphDefs.ListViewBasic'] = None

    _type: str = 'app.bsky.actor.defs#viewerState'


Preferences = List[
    Union['models.AppBskyActorDefs.AdultContentPref', 'models.AppBskyActorDefs.ContentLabelPref', 'Dict[str, Any]']
]


@dataclass
class AdultContentPref(base.ModelBase):

    """Definition model for :obj:`app.bsky.actor.defs`.

    Attributes:
        enabled: Enabled.
    """

    enabled: bool

    _type: str = 'app.bsky.actor.defs#adultContentPref'


@dataclass
class ContentLabelPref(base.ModelBase):

    """Definition model for :obj:`app.bsky.actor.defs`.

    Attributes:
        label: Label.
        visibility: Visibility.
    """

    label: str
    visibility: str

    _type: str = 'app.bsky.actor.defs#contentLabelPref'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import List, Optional

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class Data(base.DataModelBase):

    """Input data model for :obj:`com.atproto.server.createInviteCodes`.

    Attributes:
        codeCount: Code count.
        useCount: Use count.
        forAccounts: For accounts.
    """

    codeCount: int
    useCount: int
    forAccounts: Optional[List[str]] = None


@dataclass
class Response(base.ResponseModelBase):

    """Output data model for :obj:`com.atproto.server.createInviteCodes`.

    Attributes:
        codes: Codes.
    """

    codes: List['models.ComAtprotoServerCreateInviteCodes.AccountCodes']


@dataclass
class AccountCodes(base.ModelBase):

    """Definition model for :obj:`com.atproto.server.createInviteCodes`.

    Attributes:
        account: Account.
        codes: Codes.
    """

    account: str
    codes: List[str]

    _type: str = 'com.atproto.server.createInviteCodes#accountCodes'



##################################################################
# THIS IS THE AUTO-GENERATED CODE. DON'T EDIT IT BY HANDS!
# Copyright (C) 2023 Ilya (Marshal) <https://github.com/MarshalX>.
# This file is part of Python atproto SDK. Licenced under MIT.
##################################################################


from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from atproto.xrpc_client import models
from atproto.xrpc_client.models import base


@dataclass
class PostView(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        uri: Uri.
        cid: Cid.
        author: Author.
        record: Record.
        embed: Embed.
        replyCount: Reply count.
        repostCount: Repost count.
        likeCount: Like count.
        indexedAt: Indexed at.
        viewer: Viewer.
        labels: Labels.
    """

    author: 'models.AppBskyActorDefs.ProfileViewBasic'
    cid: str
    indexedAt: str
    record: 'base.RecordModelBase'
    uri: str
    embed: Optional[
        Union[
            'models.AppBskyEmbedImages.View',
            'models.AppBskyEmbedExternal.View',
            'models.AppBskyEmbedRecord.View',
            'models.AppBskyEmbedRecordWithMedia.View',
            'Dict[str, Any]',
        ]
    ] = None
    labels: Optional[List['models.ComAtprotoLabelDefs.Label']] = None
    likeCount: Optional[int] = None
    replyCount: Optional[int] = None
    repostCount: Optional[int] = None
    viewer: Optional['models.AppBskyFeedDefs.ViewerState'] = None

    _type: str = 'app.bsky.feed.defs#postView'


@dataclass
class ViewerState(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        repost: Repost.
        like: Like.
    """

    like: Optional[str] = None
    repost: Optional[str] = None

    _type: str = 'app.bsky.feed.defs#viewerState'


@dataclass
class FeedViewPost(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        post: Post.
        reply: Reply.
        reason: Reason.
    """

    post: 'models.AppBskyFeedDefs.PostView'
    reason: Optional[Union['models.AppBskyFeedDefs.ReasonRepost', 'Dict[str, Any]']] = None
    reply: Optional['models.AppBskyFeedDefs.ReplyRef'] = None

    _type: str = 'app.bsky.feed.defs#feedViewPost'


@dataclass
class ReplyRef(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        root: Root.
        parent: Parent.
    """

    parent: 'models.AppBskyFeedDefs.PostView'
    root: 'models.AppBskyFeedDefs.PostView'

    _type: str = 'app.bsky.feed.defs#replyRef'


@dataclass
class ReasonRepost(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        by: By.
        indexedAt: Indexed at.
    """

    by: 'models.AppBskyActorDefs.ProfileViewBasic'
    indexedAt: str

    _type: str = 'app.bsky.feed.defs#reasonRepost'


@dataclass
class ThreadViewPost(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        post: Post.
        parent: Parent.
        replies: Replies.
    """

    post: 'models.AppBskyFeedDefs.PostView'
    parent: Optional[
        Union[
            'models.AppBskyFeedDefs.ThreadViewPost',
            'models.AppBskyFeedDefs.NotFoundPost',
            'models.AppBskyFeedDefs.BlockedPost',
            'Dict[str, Any]',
        ]
    ] = None
    replies: Optional[
        List[
            Union[
                'models.AppBskyFeedDefs.ThreadViewPost',
                'models.AppBskyFeedDefs.NotFoundPost',
                'models.AppBskyFeedDefs.BlockedPost',
                'Dict[str, Any]',
            ]
        ]
    ] = None

    _type: str = 'app.bsky.feed.defs#threadViewPost'


@dataclass
class NotFoundPost(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        uri: Uri.
        notFound: Not found.
    """

    notFound: bool
    uri: str

    _type: str = 'app.bsky.feed.defs#notFoundPost'


@dataclass
class BlockedPost(base.ModelBase):

    """Definition model for :obj:`app.bsky.feed.defs`.

    Attributes:
        uri: Uri.
        blocked: Blocked.
    """

    blocked: bool
    uri: str

    _type: str = 'app.bsky.feed.defs#blockedPost'



from dataclasses import dataclass

from atproto.exceptions import ModelFieldNotFoundError


@dataclass
class ModelBase:
    def __getitem__(self, item):
        if hasattr(self, item):
            return getattr(self, item)

        raise ModelFieldNotFoundError(f"Can't find field '{item}' in the object of type {type(self)}.")


@dataclass
class ParamsModelBase(ModelBase):
    pass


@dataclass
class DataModelBase(ModelBase):
    pass


@dataclass
class ResponseModelBase(ModelBase):
    pass


@dataclass
class RecordModelBase(ModelBase):
    pass



import json
import os
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import dacite

from atproto.exceptions import (
    LexiconParsingError,
    UnknownDefinitionTypeError,
    UnknownPrimitiveTypeError,
)
from atproto.lexicon import models

_DEBUG = os.environ.get('DEBUG', '0') == '1'

_PATH_TO_LEXICONS = Path(__file__).parent.parent.parent.joinpath('lexicons').absolute()
_LEXICON_FILE_EXT = '.json'
_LEX_DEFINITION_TYPE_TO_CLASS = {
    models.LexDefinitionType.RECORD: models.LexRecord,
    models.LexDefinitionType.PROCEDURE: models.LexXrpcProcedure,
    models.LexDefinitionType.QUERY: models.LexXrpcQuery,
    models.LexDefinitionType.SUBSCRIPTION: models.LexSubscription,
    models.LexDefinitionType.TOKEN: models.LexToken,
    models.LexDefinitionType.OBJECT: models.LexObject,
    models.LexDefinitionType.BLOB: models.LexBlob,
    models.LexDefinitionType.ARRAY: models.LexArray,
    # TODO(MarshalX): definitions could be primitives?
    models.LexDefinitionType.STRING: models.LexString,
}

_LEX_PRIMITIVE_TYPE_TO_CLASS = {
    models.LexPrimitiveType.BOOLEAN: models.LexBoolean,
    models.LexPrimitiveType.NUMBER: models.LexNumber,
    models.LexPrimitiveType.INTEGER: models.LexInteger,
    models.LexPrimitiveType.STRING: models.LexString,
    models.LexPrimitiveType.REF: models.LexRef,
    models.LexPrimitiveType.UNION: models.LexRefUnion,
    models.LexPrimitiveType.UNKNOWN: models.LexUnknown,
    models.LexPrimitiveType.CID_LINK: models.LexCidLink,
    models.LexPrimitiveType.BYTES: models.LexBytes,
}


def _lex_definition_type_hook(data: dict) -> models.LexDefinition:
    try:
        definition_class = _LEX_DEFINITION_TYPE_TO_CLASS[models.LexDefinitionType(data['type'])]
    except (ValueError, KeyError):
        raise UnknownDefinitionTypeError('Unknown definition type ' + data['type'])

    return lexicon_parse(data, definition_class)


def _lex_primitive_type_hook(data: dict) -> models.LexPrimitive:
    try:
        primitive_class = _LEX_PRIMITIVE_TYPE_TO_CLASS[models.LexPrimitiveType(data['type'])]
    except (ValueError, KeyError):
        raise UnknownPrimitiveTypeError('Unknown primitive type ' + data['type'])

    return lexicon_parse(data, primitive_class)


_TYPE_HOOKS = {models.LexDefinition: _lex_definition_type_hook, models.LexPrimitive: _lex_primitive_type_hook}
_DEFAULT_DACITE_CONFIG = dacite.Config(cast=[Enum], type_hooks=_TYPE_HOOKS, check_types=_DEBUG)


def lexicon_parse(data: dict, data_class=None):
    if not data_class:
        data_class = models.LexiconDoc

    return dacite.from_dict(data_class=data_class, data=data, config=_DEFAULT_DACITE_CONFIG)


def lexicon_parse_file(lexicon_path: Union[Path, str], *, soft_fail: bool = False) -> Optional[models.LexiconDoc]:
    try:
        with open(lexicon_path, 'r', encoding='UTF-8') as f:
            plain_lexicon = json.loads(f.read())
            return lexicon_parse(plain_lexicon)
    except Exception as e:
        if soft_fail:
            return None

        raise LexiconParsingError("Can't parse lexicon") from e


def lexicon_parse_dir(path: Union[Path, str] = None, *, soft_fail: bool = False) -> List[models.LexiconDoc]:
    if path is None:
        path = _PATH_TO_LEXICONS

    parsed_lexicons = []

    for _, _, lexicons in os.walk(path):
        for lexicon in lexicons:
            if not lexicon.endswith(_LEXICON_FILE_EXT):
                continue

            lexicon_path = path.joinpath(lexicon)
            parsed_lexicon = lexicon_parse_file(lexicon_path, soft_fail=soft_fail)
            if parsed_lexicon:
                parsed_lexicons.append(parsed_lexicon)

    return parsed_lexicons


if __name__ == '__main__':
    lex = lexicon_parse_file(_PATH_TO_LEXICONS.joinpath('app.bsky.actor.profile.json'))
    all_lex = lexicon_parse_dir(_PATH_TO_LEXICONS)
    print('Done')
